# -*- coding: utf-8 -*-
from __future__ import division
from shm import *
from shm_updater import *
from ap_sender import Sender
from Queue import Queue
from datetime import datetime 

import random as rand
import time, sys, getopt, ast, operator

shm = Shm()

###### TO SUPPORT FIXED CHANNEL ######
FIXED = False
FIXED_CHANNEL = 1

########## DEFAULT NOT RANDOM ########
RANDOM = False

########## SLOT VARIABLES TIME #######
SLOT_TIME_SEC       = 4
SLOT_TIME           = int(SLOT_TIME_SEC * 10**6)

BEST_CH_TOTAL_SLOTS = 26 # 4 + 26
FIRST_IDLE_SLOTS    = 2
BOOTSTRAP_CYCLES    = 2 # Must be >= 2
PROBE_CHANNEL_DIST  = 4

START_CHANNEL       = 1  # Default
START_IN_SEC        = 10 # Required
START_TIME          = shm.getCurrentTime() + (START_IN_SEC * 10**6)
# Nuber empirically decided based on our observation
DIFF_THRESHOLD = 500

##### CSV FILE CONSTANTS AND SETUP ######
CSV_FILENAME   = '/tmp/dump_file.csv'

#########################################
key_read_time_usr = "CT_off_read_time_usr_64"
key_read_time_ker = "CT_off_read_time_ker_64"
key_is_probe  = "CT_off_is_probe_16"
key_is_best   = "is_best" 
key_loss_metric   = "loss_metric"
key_diff          = "diff"
key_round_number  = "round_number"
key_fhbss_packets = "fhbss_packets"
key_count_tx_tot  = "count_tx_tot"

# setup and create the csv log file
csvfile = open(CSV_FILENAME, 'w')
# take the keys sorted by offset to write the fieldnames
fieldnames = [key for key,value in sorted(OFFSETS.items(), key=operator.itemgetter(1))]
fieldnames += [key_read_time_ker, key_read_time_usr, key_is_best, key_loss_metric, 
        key_diff, key_round_number, key_fhbss_packets, key_count_tx_tot]

writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
writer.writerow(dict((f, f if 'CT_off' not in f else f[7:-3]) for f in fieldnames))
############################################

def get_bootstrap_switches(last_sw_time):
    if not FIXED:
        # Get random bootstrap switches
        channels = range(1,14)
        rand.shuffle(channels)
    else:
        channels = [FIXED_CHANNEL] * 13

    return _get_bootstrap_switches(channels, last_sw_time)
    
def _get_bootstrap_switches(channels, last_sw_time):
    # Setup variables
    sw_time  = int(last_sw_time, 16)
    switches = []

    print "sw_time = {0}, last_sw_time = {1}".format(sw_time, last_sw_time)

    for channel in channels:
        sw_time += SLOT_TIME
        switches += [ (channel, hex(sw_time), 1) ] # 1 = is_probe
    
    # dbg print
    print "BOOTSTRAP SWITCHES"
    for (a,b,c) in switches:
        print "ch={0:2}, is_probe={2}, time={1}".format(a,b,c)

    # next_switches, probe_total_slots, best_total_slots 
    return switches, len(switches), 0

def add_sw_times(l, last_switch_time):
     # setup the offset
    offset = 0
    # setup the final list with sw_times
    for i, el in enumerate(l):
        offset += SLOT_TIME_SEC
        (ch, is_probe) = l[i]
        l[i] = (ch, hex(long(last_switch_time, 16) + int(offset * 1000000)), is_probe)

def next_cycle_switches_fixed(last_switch_time):
    l = [(FIXED_CHANNEL, 0)] * BEST_CH_TOTAL_SLOTS
    # Add fake probe
    l.extend([(FIXED_CHANNEL, 1)] * 13)
    rand.shuffle(l)
    add_sw_times(l, last_switch_time)
    # next_switches, probe_total_slots, best_total_slots 
    return l, 13, BEST_CH_TOTAL_SLOTS


def next_cycle_switches(best_channel_number, last_switch_time, busy_channels):

    REMAINING_SLOTS = BEST_CH_TOTAL_SLOTS - FIRST_IDLE_SLOTS
    # create idle slot list [(1,0), (1,0), ...]
    l = [(best_channel_number, 0)] * FIRST_IDLE_SLOTS
    # create remaining slots list for best channel
    remaining = [(best_channel_number, 0)] * REMAINING_SLOTS
    # Calculate smart probing slots
    probe_channels = range(1,14)
    ########################## SMART PROBING ###################################
    # if not RANDOM:
    for n in busy_channels:
        channels_with_interferences = [i for i in range(n-PROBE_CHANNEL_DIST,n+PROBE_CHANNEL_DIST+1) if i > 0 and i < 14 and i != n]
        probe_channels = list(set(probe_channels) - set(channels_with_interferences))
    probe_channels = list(set(busy_channels + probe_channels))
    ############################################################################
    remaining.extend([(n, 1) for n in probe_channels])
    # shuffle not idle slots list
    rand.shuffle(remaining)
    # put toghether the two lists
    l.extend(remaining)
    # setup the final list with sw_times
    add_sw_times(l, last_switch_time)
    # next_switches, probe_total_slots, best_total_slots 
    return l, len(probe_channels), BEST_CH_TOTAL_SLOTS

# Called every round 
def calculate_metrics(data, round_number):
    busy_channels = []
    best_channel_number = 1 ## for the fucking first slot demmerda
    # Add metrics to each row and save the busy channels
    for row_data in data:
        (ch, diff, is_probe) = add_metrics(row_data, round_number)
        if is_probe and diff > DIFF_THRESHOLD:
            busy_channels += [ch]

    ######################### CHANNEL CHOICE ###########################
    # Only between the probes calculate the best_channel
    probes = [row for row in data if row[key_is_probe]]
    if probes: # This if is for the motherfucker 1 slot at the beginning
        three_best = sorted(probes,  key=lambda row: row[key_loss_metric])[:3]
        best_row   = min(three_best,  key=lambda row: row[key_diff])
        # FLAG the best channel
        best_row[key_is_best] = 1
        best_channel_number = best_row["CT_off_channel_number_16"]
    ####################################################################
    writer.writerows(data)
    csvfile.flush()

    return best_channel_number, busy_channels


def add_metrics(row_data, round_number):

    count_rx_succ_to_me = row_data['CT_off_count_rx_succ_to_me_32']
    count_rx_succ_tot = row_data['CT_off_count_rx_succ_tot_32']
    channel_number = row_data['CT_off_channel_number_16']
    count_rx_error = row_data['CT_off_count_rx_error_32']
    count_tx_lost = row_data['CT_off_count_tx_lost_32']
    count_tx_succ = row_data['CT_off_count_tx_32']
    rx_bad_plcp = row_data['CT_off_rx_bad_plcp_32']

    ############## METRIC EVALUATION ####################
    errors = count_rx_error + count_tx_lost + rx_bad_plcp
    successes =  count_tx_succ + count_rx_succ_to_me
    # Mai capitato che non trasmette neanche un pacchetto boh assurdo
    if errors + successes > 0:
        # LOSS METRIC => evaluate best channel
        loss_metric = errors / (errors + successes)
    else:
        loss_metric = 1
    # DIFF => check channel occupation by others
    diff = count_rx_succ_tot - count_rx_succ_to_me
    #####################################################
    
    # Save the calculated values
    row_data[key_loss_metric] = loss_metric
    row_data[key_diff] = diff
    row_data[key_is_best] = 0
    row_data[key_round_number] = round_number
    # Estimate of the throughput
    row_data[key_fhbss_packets] = count_rx_succ_to_me + count_tx_succ            
    # Packet tx tot
    row_data[key_count_tx_tot] = count_tx_succ + count_tx_lost
    
    is_probe = row_data[key_is_probe]
    # DBG PRINT
    if is_probe:
        print "channel=%s, loss_metric=%f, diff=%d" % (channel_number, loss_metric, diff)
    # Return this to calculate easy the busy channels
    return (channel_number, diff, is_probe)


########## START FHBSS ############
opts, args = getopt.getopt(sys.argv[1:], '', ['random', 'start-channel=', 'fixed=', 'stations='])
for o, a in opts:
    if o == '--random':
        print "FHBSS RANDOM"
        RANDOM=True

    elif o == '--stations':
        STA_NUMBERS = ast.literal_eval(a)

    elif o == '--start-channel':
        START_CHANNEL = int(a)

    elif o in ('--fixed'):
        print "FHBSS FIXED ON CHANNEL =", a
        FIXED = True
        FIXED_CHANNEL = int(a)

if not STA_NUMBERS:
    raise Exception("Please set the station numbers using the --stations option!")

shm_switch_queue = Queue()
sender_queue = Queue()
data_queue = Queue()

sender = Sender(sender_queue, STA_NUMBERS)
shm_updater = APShmUpdater(shm, shm_switch_queue, data_queue)

############################################################################
# NB: Remember that by design, also of the firmware, the first bootstrap slot  
#     MUST be in channel 1 and a NON PROBING slot (because it sets the start time)
switches = [ (START_CHANNEL, hex(START_TIME), 0) ]
############################################################################
next_switches, probe_total_slots, best_total_slots = get_bootstrap_switches(hex(START_TIME))

switches += next_switches
last_switch_time = switches[-1][1]

sender_queue.put(switches)
shm_updater.bootstrap(switches)

data = []
metrics = []
round_number = 0
probe_count = 0
best_count  = 0
prev_probe_total_slots = 0
prev_best_total_slots = 1 # => because the first slot is a non probe in channel 1

######## Ora Calcola i next switches ############
while True:
    
    # Check if a cycle is finished
    if probe_count == prev_probe_total_slots and best_count == prev_best_total_slots:
        # salvare il csv calcolare metrica, best e busy
        best, busy = calculate_metrics(data, round_number)
        # dbg print
        print "round=%s, best_channel=%s" % (round_number, best)

        # Support for more then 1 bootstrap cycles, and if random continue to go randomly
        if RANDOM or round_number < BOOTSTRAP_CYCLES - 1:
            next_switches, next_probe_total_slots, next_best_total_slots = \
                    get_bootstrap_switches(last_switch_time)
        else:
            if FIXED:
                next_switches, next_probe_total_slots, next_best_total_slots = \
                        next_cycle_switches_fixed(last_switch_time)
            else:
                next_switches, next_probe_total_slots, next_best_total_slots = \
                    next_cycle_switches(best, last_switch_time, busy)

        # Update the prev
        prev_probe_total_slots = probe_total_slots
        prev_best_total_slots = best_total_slots
        # now update the next
        probe_total_slots = next_probe_total_slots
        best_total_slots = next_best_total_slots
        # update last_switch_time
        last_switch_time = next_switches[-1][1]
        # send the switches to the STAs
        sender_queue.put(next_switches)
        # put the last switch in the queue
        for (channel, sw_time, is_probe) in next_switches:
            shm_switch_queue.put((channel, sw_time, is_probe))
        # once calculate everything update counters
        round_number += 1
        probe_count = 0 
        best_count = 0
        # and free memory
        del data[:]

    sys.stdout.flush()
    # Wait to get new data from the SHM
    row_data = data_queue.get()
    # Save the data temporary
    data += [row_data]
    # check if i get a probe
    if row_data[key_is_probe]: 
        probe_count += 1
    else:
        best_count += 1
