from libb43 import *
from sys import stdout

ADDRESSES = {

    'time_now_w3'       : 0xFD0,
    'time_now_w2'       : 0xFD2,
    'time_now_w1'       : 0xFD4,
    'time_now_w0'       : 0xFD6,

    'current_row'       : 0xFFA,
    'is_ct_initialized' : 0xFFC,
    'is_fh_active'      : 0xFFE
}

CT = {
    'CT_base'       : 0xF10,
    'CT_eof'        : 0xFB0,
    'CT_entry_size' : 0x50
}

OFFSETS = {

    'CT_off_channel_code_16'        : 0,
    'CT_off_channel_number_16'      : 0x02,
    'CT_off_is_probe_16'            : 0x04,
    'CT_off_switching_time_64'      : 0x06,
    'CT_off_current_line_dbg_16'    : 0x0E,
    
    'CT_off_time_rx_succ_to_me_32'  : 0x10,
    'CT_off_time_rx_succ_tot_32'    : 0x14,
    'CT_off_time_rx_error_32'       : 0x18,
    'CT_off_time_tx_32'             : 0x1C,

    'CT_off_count_rx_succ_to_me_32' : 0x20,
    'CT_off_count_rx_succ_tot_32'   : 0x24,
    'CT_off_count_rx_error_32'      : 0x28,
    'CT_off_count_tx_32'            : 0x2C,

    'CT_off_count_tx_lost_32'       : 0x30,
    'CT_off_access_time_32'         : 0x34,
    'CT_off_cs_busy_time_32'        : 0x38,
    'CT_off_cs_busy_time_16ms_32'   : 0x3C,

    'CT_off_rx_bad_plcp_32'         : 0x40,
    'CT_off_access_time_ma_32'      : 0x44,
    
}

# ------------------------------------------------------
# NB: dall'assembly scriviamo sempre i numeri, contatori 
# su word da 16 con le piu significative a SINISTRA
# ------------------------------------------------------

class Shm(object):
    def __init__(self):
        super(Shm, self).__init__()
        self.b43 = B43()

    def _read16(self, offset):
        return self.b43.shmRead16(1, offset)

    def _read32(self, offset):
        w1 = self._read16(offset) << 16
        w0 = self._read16(offset + 2)
        return w1 + w0

    def _read64(self, offset):
        w1 = self._read32(offset) << 32
        w0 = self._read32(offset + 4)
        return w1 + w0

    def _write16(self, offset, value_16):
        self.b43.shmWrite16(1, offset, value_16)

    def _write32(self, offset, value_32):
        self.b43.shmWrite16(1, offset, value_32 >> 16)
        self.b43.shmWrite16(1, offset + 2, value_32 - ((value_32 >> 16) << 16))

    def _write64(self, offset, value_64):
        self._write32(offset, value_64 >> 32)
        self._write32(offset + 4, value_64 - ((value_64 >> 32) << 32))

    def read16(self, key, row):
        return self._read16(self._getOffset_(key, row))

    def read32(self, key, row):
        return self._read32(self._getOffset_(key, row))

    def read64(self, key, row):
        return self._read64(self._getOffset_(key, row))

    def write16(self, key, row, value):
        self._write16(self._getOffset_(key, row), value)

    def write32(self, key, row, value):
        self._write32(self._getOffset_(key, row), value)

    def _getOffset_(self, key, row):
        if row not in [0, 1]:
            raise Exception("Row must be 0 or 1.")
        return CT['CT_base'] + ( row * CT['CT_entry_size'] ) + OFFSETS[key]

    def writeSwitch(self, row, channel_number, sw_time_64, is_probe=0):
        if isinstance(sw_time_64, basestring):
            sw_time_64 = long(sw_time_64, 16)
        # write the time
        self._write64(self._getOffset_('CT_off_switching_time_64', row), sw_time_64)
        # write channel code and number
        self._write16(self._getOffset_('CT_off_channel_code_16', row), self.getChannelCode(channel_number))
        self._write16(self._getOffset_('CT_off_channel_number_16', row), channel_number)
        # write is_probe
        self._write16(self._getOffset_('CT_off_is_probe_16', row), is_probe)

    def getCurrentTime(self):
        return self._read64(ADDRESSES['time_now_w3'])

    def activateFhbss(self):
        self._write16(ADDRESSES['is_fh_active'], 1)

    def getChannelCode(self, channelNumber):
        if channelNumber == 14:
            return 0x8054
        else: 
            code = 0x800C
            for i in range(1, channelNumber):
                code += 0x5
            return code
