# -*- coding: utf-8 -*-
from Queue import Empty
from thread import *
from shm import ADDRESSES, OFFSETS

import time, re, csv

class ShmUpdater(object):

    def __init__(self, shm, sw_queue):
        super(ShmUpdater, self).__init__()
        self.sw_queue = sw_queue
        self.shm = shm
        self.old_value = self.shm._read16(ADDRESSES['current_row'])
        self.log_file = open('/tmp/shm_updater.log', 'a')

    def log(self, string):
        self.log_file.write(string)
        self.log_file.flush()

    def time_val(self, hex_time)
        return long(hex_time, 16) / 1000000

    def is_switched(self):
        # Return a boolean that is true if detect that the station is switched, waiting before 1 second
        time.sleep(1.5) ## < SLOT_TIME

        current_value = self.shm._read16(ADDRESSES['current_row'])
        if self.old_value != current_value:
            self.old_value = current_value
            # it's possible to write in line 1 or 0 depends from current_value
            return (True, 1) if current_value == 1928 else (True, 0)

        return (False, -1)

    def bootstrap(self, switches):
        if len(switches) < 2:
            raise Exception("For bootstrap 2 switches are required!")
        # get the first 2 and write it to the first line
        for index in [0, 1]:
            self.shm.writeSwitch(index, *switches[index])
        # metto in coda dal terzo in poi because il secondo l'ho gia' scritto prima
        for switch in switches[2:]: 
            self.sw_queue.put(switch)
        # start the thread for updating the shm
        start_new_thread( self.shm_updater, ())
        # activate the FHBSS firmware
        self.shm.activateFhbss()

    def shm_updater(self):
        log_template = "{0:16} row = {1:2}, channel = {2:2}, sw_time = {3:6}\n"

        # infinite loop
        while True:
            (has_switched, row_index) = self.is_switched()
            # when it understand that has switched
            if has_switched:
                try: # try to get the new switch to write
                    (channel, sw_time) = self.sw_queue.get(False)
                except Empty, e:
                    error = "No more scheduled switches. 3..2..1.. EXPLODE!!"
                    self.log(error)
                    raise RuntimeError(error)
                else:
                    # and then update the next channel and time
                    self.shm.writeSwitch(row_index, channel, sw_time)
                    self.log(log_template.format(str(time.time()), row_index, channel, self.time_val(sw_time)))

class APShmUpdater(ShmUpdater):
    
    def __init__(self, shm, sw_queue, data_queue): # time mapper non serve piu era solo per le stampe
        super(APShmUpdater, self).__init__(shm, sw_queue)
        self.data_queue = data_queue

    def get_data(self, row):
        # Get the data written in the SHM by the firmware
        data = {}
        for key in OFFSETS.keys():
            if re.search("_16$", key): 
                data[ key ] = self.shm.read16(key, row)
            elif re.search("_32$", key):
                data[ key ] = self.shm.read32(key, row)
            elif re.search("_64$", key):
                data[ key ] = self.shm.read64(key, row)
        return data

    def shm_updater(self):
        log_template = "{0:16} row = {1:2}, channel = {2:2}, sw_time = {3:6}, is_probe = {4:2}\n"
        
        # start effectively the loop
        while True:
            (has_switched, row_index) = self.is_switched()
            # when it understand that has switched
            if has_switched:
                
                data = self.get_data(row_index)
                
                data["CT_off_read_time_ker_64"] = self.shm.getCurrentTime()
                data["CT_off_read_time_usr_64"] = time.time() # insert also the user time
                
                self.data_queue.put(data)

                try: # try to get the new switch to write
                    (channel, sw_time, is_probe) = self.sw_queue.get(False)
                except Empty, e:
                    error = "%s No more scheduled switches. 3..2..1.. EXPLODE!!" % time.time()
                    self.log_file.write(error)
                    self.log_file.close()
                    raise RuntimeError(error)
                else:
                    # update the row with the next channel and time
                    self.shm.writeSwitch(row_index, channel, sw_time, is_probe)
                    self.log(log_template.format(str(time.time()), row_index, channel, self.time_val(sw_time), is_probe))
